// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include "InternalFrame.h"
#include "InternalSideData.h"
#include <cstring>

namespace xma {

#define XMA_FRAME "Frame"
#define ROI_SB_INDEX 0

const char* pixel_format_names[] = {"none", "nv12", "nv21", "p010le", "p010be", "yuv420p", "yuv420p10le", "yuv420p10be", "yuv422p", "yuv422p10le", "yuv422p10be",
    "yuv444p", "argb", "abgr", "rgba", "bgra", "rgb24", "bgr24", "vpe", "packed10", "bgr0", "rgb16", "rgbp", "bgrp", "rgb48le", "bgr48le", "rgb48p", "bgr48p"};

//////////////////////////////////////////////////////////////////////////////
PrivateFrameInfo::PrivateFrameInfo(IntSessionFactoryPtr sf, LogPtr log, buffer_t buffer) : m_buffer(std::move(buffer)), m_sf(sf), m_log(log) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateFrameInfo::PrivateFrameInfo()");
}

PrivateFrameInfo::~PrivateFrameInfo() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateFrameInfo::~PrivateFrameInfo()");
  if (m_pic) {
    m_log->LogMsg(XMA_WARNING_LOG, XMA_FRAME, "PrivateFrameInfo::~PrivateFrameInfo() VpiFrm not empty!");
  }
  if (m_parent) {
    m_parent->DecRef();
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateFrameInfo::~PrivateFrameInfo() - returning");
}

bool PrivateFrameInfo::IsValid() const {
  return (m_dummy == FRAME_DUMMY_VALUE);
}

std::recursive_mutex& PrivateFrameInfo::GetMutex() {
  return m_mutex;
}

VpiBufferPool* PrivateFrameInfo::GetBufferPool() {
  if (m_buffer) {
    return static_cast<VpiBufferPool*>(m_buffer->pool);
  } else if (m_pic) {
    VpiBuffer* buf = static_cast<VpiBuffer*>(m_pic->vpi_picinfo_buf);
    if (buf) {
      return static_cast<VpiBufferPool*>(buf->pool);
    }
  }
  if (m_parent) {
    IntFramePtr parent = std::dynamic_pointer_cast<InternalFrame>(m_parent);
    if (parent) {
      return parent->GetBufferPool();
    }
  }
  return nullptr;
}

void PrivateFrameInfo::SetSelf(IntFramePtr self) {
  m_self = self;
}

IntFramePtr PrivateFrameInfo::GetSelf() const {
  return m_self;
}

void PrivateFrameInfo::SetParent(FramePtr parent) {
  m_parent = parent;
  m_parent->IncRef();
}

int32_t PrivateFrameInfo::AddVpiFrm(VpiFrm* pic) {
  m_pic = vpi_frm_t(pic, [&](VpiFrm* frm) {
    int ref_count = vpi_buffer_get_ref_count(static_cast<VpiBuffer*>(frm->vpi_picinfo_buf));
    vpi_frame_unref(frm);
    if (ref_count == 1) {
      vpi_frame_free(m_sf->GetMemHandle(), frm);
    }
  });
  return XMA_SUCCESS;
}

VpiFrm* PrivateFrameInfo::GetVpiFrm() const {
  return m_pic.get();
}

int32_t PrivateFrameInfo::RemoveVpiFrm() {
  int32_t ret = m_pic ? XMA_SUCCESS : XMA_ERROR;
  m_pic.reset();
  return ret;
}

void PrivateFrameInfo::SetCallback(std::function<void(uint8_t**)> free_callback) {
  m_free_callback = free_callback;
}

void PrivateFrameInfo::CallCallback(uint8_t** frame_data) {
  if (m_free_callback) {
    m_free_callback(frame_data);
  }
}

void PrivateFrameInfo::SetFrozen() {
  m_frozen = true;
}

void PrivateFrameInfo::Unfreeze() {
  m_frozen = false;
}

bool PrivateFrameInfo::IsFrozen() const {
  return m_frozen;
}

IntSessionFactoryPtr PrivateFrameInfo::GetSessionFactory() {
  return m_sf;
}

LogPtr PrivateFrameInfo::GetLog() {
  return m_log;
}

//////////////////////////////////////////////////////////////////////////////
PrivateVpiFrmInfo::PrivateVpiFrmInfo(IntSessionFactoryPtr sf, LogPtr log) : m_sf(sf), m_log(log) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::PrivateVpiFrmInfo()");
}

PrivateVpiFrmInfo::~PrivateVpiFrmInfo() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::~PrivateVpiFrmInfo()");
  if (!m_frames.empty()) {
    m_log->LogMsg(XMA_WARNING_LOG, XMA_FRAME, "PrivateVpiFrmInfo::~PrivateVpiFrmInfo() Frame list not empty!");
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::~PrivateVpiFrmInfo() - returning");
}

bool PrivateVpiFrmInfo::IsValid() const {
  return (m_dummy == VPIFRM_DUMMY_VALUE);
}

int32_t PrivateVpiFrmInfo::AddFrame(FramePtr frame, VpiFrm* pic, int32_t ch) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::AddFrame(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(frame->GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(frame->GetProps()->GetSwFormat()),
      frame->GetProps()->GetWidth(), frame->GetProps()->GetHeight());
  vpi_frame_info_t vpi_info = std::make_unique<VpiFrmInfo>();
  vpi_info->frame           = frame;
  vpi_info->buf             = static_cast<VpiBuffer*>(pic->vpi_picinfo_buf);
  int32_t    planes         = frame->GetProps()->GetNumPlanes();
  int32_t    plane;
  VpiBuffer* buf;
  switch (frame->GetBufferType()) {
  case XMA_HOST_BUFFER_TYPE:
  case XMA_DEVICE_ONLY_BUFFER_TYPE:
  case XMA_DEVICE_BUFFER_TYPE: {
    buf                    = reinterpret_cast<VpiBuffer*>(pic->vpi_picinfo_buf);
    FramePtr    iter_frame = frame;
    IntFramePtr int_frame  = std::dynamic_pointer_cast<InternalFrame>(frame);
    if (int_frame) {
      XmaFrame* xma_frame = int_frame->GetFrame();
      if (xma_frame) {
        PrivateFrameInfo* frame_info = static_cast<PrivateFrameInfo*>(xma_frame->private_do_not_touch);
        if (frame_info && frame_info->IsValid()) {
          VpiFrm* frm = frame_info->GetVpiFrm();
          if (frm) {
            int32_t frm_ch;
            for (frm_ch = 0; frm_ch < VPE_CHANNEL_NUM; ++frm_ch) {
              if ((frm->ch_info[frm_ch].enabled) && (frm->ch_info[frm_ch].format == pic->ch_info[ch].format) &&
                  (frm->ch_info[frm_ch].sw_format == pic->ch_info[ch].sw_format) && (frm->ch_info[frm_ch].width == pic->ch_info[ch].width) &&
                  (frm->ch_info[frm_ch].height == pic->ch_info[ch].height)) {
                break;
              }
            }
            if (frm_ch < VPE_CHANNEL_NUM) {
              VpiBuffer* frm_buf     = static_cast<VpiBuffer*>(frm->vpi_picinfo_buf);
              buf->m_handle          = frm_buf->m_handle;
              buf->m_metadata_handle = frm_buf->m_metadata_handle;
              memcpy(buf->buf_paddr[ch], frm_buf->buf_paddr[frm_ch], sizeof(void*) * VPE_MAX_PLANES);
              memcpy(buf->buf_vaddr[ch], frm_buf->buf_vaddr[frm_ch], sizeof(void*) * VPE_MAX_PLANES);
              memcpy(buf->md_paddr[ch], frm_buf->md_paddr[frm_ch], sizeof(void*) * VPE_MAX_PLANES);
              memcpy(buf->md_vaddr[ch], frm_buf->md_vaddr[frm_ch], sizeof(void*) * VPE_MAX_PLANES);
              buf->buf_size     = frm_buf->buf_size;
              buf->md_data_size = frm_buf->md_data_size;
              m_frames.push_back(std::move(vpi_info));
              m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::AddFrame() - Returning success");
              return XMA_SUCCESS;
            }
          }
        }
      }
    } else {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "PrivateVpiFrmInfo::AddFrame() invalid frame!");
    }
    buf->m_handle = static_cast<VpiLinearMem*>(frame->GetBuffer(0));
    for (plane = 0; plane < planes; plane++) {
      VpiLinearMem* mem         = static_cast<VpiLinearMem*>(frame->GetBuffer(plane));
      buf->buf_paddr[ch][plane] = reinterpret_cast<void*>(mem->bus_address_ep);
      buf->buf_vaddr[ch][plane] = mem->virtual_address_rc;
    }
    break;
  }
  case NO_BUFFER:
    break;
  }
  m_frames.push_back(std::move(vpi_info));
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::AddFrame() - Returning success");
  return XMA_SUCCESS;
}

vpi_frame_info_t PrivateVpiFrmInfo::GetFrame(XmaFormatType format, XmaFormatType sw_format, int32_t width, int32_t height) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::GetFrame(format=%s, sw_format=%s, size=%dx%d)", InternalFrameProperties::GetPixFmtName(format),
      InternalFrameProperties::GetPixFmtName(sw_format), width, height);
  auto itr = m_frames.begin();
  while (itr != m_frames.end()) {
    FramePropertiesPtr props = (*itr)->frame->GetProps();
    if ((props->GetFormat() == format) && (props->GetSwFormat() == sw_format) && (props->GetWidth() == width) && (props->GetHeight() == height)) {
      vpi_frame_info_t vpi_info = std::move(*itr);
      m_frames.erase(itr);
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::GetFrame() - Returning frame");
      return vpi_info;
    }
    ++itr;
  }
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::GetFrame() - Returning frame not found");
  return nullptr;
}

int32_t PrivateVpiFrmInfo::GetFrameCount() const {
  return m_frames.size();
}

int32_t PrivateVpiFrmInfo::SetFreeBuf(VpePicFree func_free) {
  m_func_free = func_free;
  return XMA_SUCCESS;
}

VpePicFree PrivateVpiFrmInfo::GetFreeBuf() const {
  return m_func_free;
}

int32_t PrivateVpiFrmInfo::SetFreeData(void* data) {
  m_free_data = data;
  return XMA_SUCCESS;
}

void* PrivateVpiFrmInfo::GetFreeData() const {
  return m_free_data;
}

int32_t PrivateVpiFrmInfo::SetPool(VpiBufferPool* pool) {
  m_pool = pool;
  return XMA_SUCCESS;
}

VpiBufferPool* PrivateVpiFrmInfo::GetPool() const {
  return m_pool;
}

IntSessionFactoryPtr PrivateVpiFrmInfo::GetSessionFactory() {
  return m_sf;
}

LogPtr PrivateVpiFrmInfo::GetLog() {
  return m_log;
}

void InitVpiFrm(VpiFrm* pic) {
  pic->valid     = 0;
  pic->width     = 0;
  pic->height    = 0;
  pic->poc       = 0;
  pic->format    = VPI_FORMAT_NONE;
  pic->key_frame = 0;
  memset(pic->vpi_sidedata, 0, sizeof(pic->vpi_sidedata));
  memset(pic->ch_info, 0, sizeof(pic->ch_info));
  pic->pts      = 0;
  pic->dts      = 0;
  pic->next     = nullptr;
  pic->opaque   = nullptr;
  pic->ch_index = 0;
}

//////////////////////////////////////////////////////////////////////////////
void xma_pic_consume(void*, void* data, void* buf) {
  VpiFrm* pic = static_cast<VpiFrm*>(data);

  if (!pic) {
    throw InvalidArgumentException("xma_pic_consume() Invalid parameter!");
  }

  PrivateVpiFrmInfo* info = static_cast<PrivateVpiFrmInfo*>(pic->opaque);

  if (!info || !info->IsValid()) {
    throw InvalidArgumentException("xma_pic_consume() Invalid parameter!");
  }

  LogPtr log = info->GetLog();
  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "xma_pic_consume()");

  VpiBuffer*     buffer  = static_cast<VpiBuffer*>(buf);
  VpiBufferPool* vpiPool = static_cast<VpiBufferPool*>(buffer->pool);
  BufferPool*    pool    = static_cast<BufferPool*>(vpiPool->opaque);
  if (pool) {
    pool->FreeBuffer(buffer);
  }
  vpi_memory_free_frame(info->GetSessionFactory()->GetMemHandle(), pic);
  pic->vpi_picinfo_buf = nullptr;

  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "xma_pic_consume() - Returning");
}

//////////////////////////////////////////////////////////////////////////////
void xma_free_reference_info(void*, void* data, void* buff) {
  VpiFrm* pic = static_cast<VpiFrm*>(data);

  if (!pic) {
    throw InvalidArgumentException("xma_free_reference_info() Invalid parameter!");
  }

  PrivateVpiFrmInfo* info = static_cast<PrivateVpiFrmInfo*>(pic->opaque);

  if (!info || !info->IsValid()) {
    throw InvalidArgumentException("xma_free_reference_info() Invalid parameter!");
  }

  LogPtr log = info->GetLog();
  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "xma_free_reference_info()");
  IntSessionFactoryPtr sf = info->GetSessionFactory();

  VpiBuffer* buf = static_cast<VpiBuffer*>(buff);
  if (buf) {
    buf->pool    = info->GetPool();
    buf->freebuf = info->GetFreeBuf();
    if (buf->freebuf) {
      buf->freebuf(buf->opaque, buf->data, buf);
    }
  }

  for (int32_t ch = 0; ch < VPE_CHANNEL_NUM; ++ch) {
    VpiChannelInfo& ch_info = pic->ch_info[ch];
    if (ch_info.enabled) {
      vpi_frame_info_t vpi_info = info->GetFrame(
          static_cast<XmaFormatType>(ch_info.format), static_cast<XmaFormatType>(ch_info.sw_format), ch_info.width, ch_info.height);
      if (vpi_info) {
        pic->vpi_picinfo_buf = vpi_info->buf;
      }
      memset(&ch_info, 0, sizeof(VpiChannelInfo));
    }
  }

  delete info;

  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "xma_free_reference_info() - Returning");
}

//////////////////////////////////////////////////////////////////////////////
InternalFrameProperties::InternalFrameProperties(
    IntSessionFactoryPtr sf, LogPtr log, enum XmaFormatType format, enum XmaFormatType sw_format, int32_t width, int32_t height, uint32_t flags) :
    m_sf(sf),
    m_log(log) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrameProperties::InternalFrameProperties(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(format), InternalFrameProperties::GetPixFmtName(sw_format), width, height);
  memset(&m_props, 0, sizeof(XmaFrameProperties));
  m_props.format    = format;
  m_props.sw_format = sw_format;
  m_props.width     = width;
  m_props.height    = height;
  m_props.flags     = flags;

  int32_t planes = GetNumPlanes();
  int32_t plane;
  for (plane = 0; plane < planes; ++plane) {
    m_props.linesize[plane] = vpi_frame_get_plane_stride(m_props.sw_format, plane, m_props.width, XMA_FRAME_W_ALIGN);
  }
  for (; plane < XMA_MAX_PLANES; ++plane) {
    m_props.linesize[plane] = 0;
  }
}

///////////////////////////////////////////////////////////////////////////////
InternalFrameProperties::InternalFrameProperties(IntSessionFactoryPtr sf, LogPtr log, XmaFrameProperties* props) : m_props(*props), m_sf(sf), m_log(log) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrameProperties::InternalFrameProperties(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(props->format), InternalFrameProperties::GetPixFmtName(props->sw_format), props->width, props->height);
  int32_t planes = GetNumPlanes();
  int32_t plane;
  for (plane = 0; plane < planes; ++plane) {
    if (m_props.linesize[plane] == 0) {
      m_props.linesize[plane] = vpi_frame_get_plane_stride(m_props.sw_format, plane, m_props.width, XMA_FRAME_W_ALIGN);
    }
  }
  for (; plane < XMA_MAX_PLANES; ++plane) {
    m_props.linesize[plane] = 0;
  }
}

///////////////////////////////////////////////////////////////////////////////
XmaFrameProperties InternalFrameProperties::GetFrameProperties() const {
  return m_props;
}

///////////////////////////////////////////////////////////////////////////////
char* InternalFrameProperties::GetPixFmtName(XmaFormatType fmt) {
  return const_cast<char*>(pixel_format_names[fmt]);
}

///////////////////////////////////////////////////////////////////////////////
enum XmaFormatType InternalFrameProperties::GetFormat() const {
  return m_props.format;
}

///////////////////////////////////////////////////////////////////////////////
enum XmaFormatType InternalFrameProperties::GetSwFormat() const {
  return m_props.sw_format;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrameProperties::GetWidth() const {
  return m_props.width;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrameProperties::GetHeight() const {
  return m_props.height;
}

///////////////////////////////////////////////////////////////////////////////
uint32_t InternalFrameProperties::GetFlags() const {
  return m_props.flags;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrameProperties::GetNumPlanes() const {
  return vpi_frame_get_planes(m_props.sw_format);
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrameProperties::GetPlaneHeight(int32_t plane) const {
  if ((plane < 0) || (plane > GetNumPlanes())) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrameProperties::GetPlaneHeight() Invalid parameter!");
  }
  return vpi_frame_get_plane_height(m_props.sw_format, plane, m_props.height, XMA_FRAME_H_ALIGN);
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrameProperties::GetPlaneStride(int32_t plane) const {
  if ((plane < 0) || (plane > GetNumPlanes())) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrameProperties::GetPlaneStride() Invalid parameter!");
  }
  return m_props.linesize[plane];
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrameProperties::GetPlaneSize(int32_t plane) const {
  if ((plane < 0) || (plane > GetNumPlanes())) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrameProperties::GetPlaneSize() Invalid parameter!");
  }
  return m_props.linesize[plane] * GetPlaneHeight(plane);
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrameProperties::GetBufferSize(const FramePropertiesPtr props) {
  int32_t planes = props->GetNumPlanes();
  int32_t size   = 1024; // Needed for AV1?
  for (int32_t plane = 0; plane < planes; ++plane) {
    size += props->GetPlaneSize(plane);
  }
  return size;
}

///////////////////////////////////////////////////////////////////////////////
InternalFrame::InternalFrame(
    IntSessionFactoryPtr sf, LogPtr log, FramePropertiesPtr props, bool dummy, buffer_t buffer, XmaFraction frame_rate, int64_t pts, bool is_idr) :
    m_props(props),
    m_sf(sf), m_log(log) {
  if (dummy) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame(dummy)");
  } else {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame(format=%s, sw_format=%s, size=%dx%d)",
        InternalFrameProperties::GetPixFmtName(props->GetFormat()), InternalFrameProperties::GetPixFmtName(props->GetSwFormat()), props->GetWidth(), props->GetHeight());
    if (!buffer) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::InternalFrame() No pool provided!");
    }
  }
  m_frame = std::make_unique<XmaFrame>();

  memset(m_frame.get(), 0, sizeof(XmaFrame));

  m_frame->frame_props.format    = props->GetFormat();
  m_frame->frame_props.sw_format = props->GetSwFormat();
  m_frame->frame_props.width     = props->GetWidth();
  m_frame->frame_props.height    = props->GetHeight();
  m_frame->frame_props.flags     = props->GetFlags();
  m_frame->frame_rate            = frame_rate;
  m_frame->pts                   = pts;
  m_frame->is_idr                = is_idr ? 1 : 0;

  int32_t planes = props->GetNumPlanes();
  int32_t plane;
  if (dummy) {
    for (plane = 0; plane < planes; ++plane) {
      m_frame->frame_props.linesize[plane] = props->GetPlaneStride(plane);
      m_frame->data[plane].buffer_type     = NO_BUFFER;
      m_frame->data[plane].buffer          = nullptr;
      m_frame->data[plane].host            = nullptr;
      m_frame->data[plane].device          = nullptr;
    }
    for (; plane < XMA_MAX_PLANES; ++plane) {
      m_frame->frame_props.linesize[plane] = 0;
      m_frame->data[plane].buffer_type     = NO_BUFFER;
      m_frame->data[plane].buffer          = nullptr;
      m_frame->data[plane].host            = nullptr;
      m_frame->data[plane].device          = nullptr;
    }
  } else {
    int32_t       sizes[XMA_MAX_PLANES];
    VpiLinearMem* mem;
    VpiLinearMem* mem0;
    for (plane = 0; plane < planes; ++plane) {
      sizes[plane]                         = props->GetPlaneSize(plane);
      m_frame->frame_props.linesize[plane] = props->GetPlaneStride(plane);
      m_frame->data[plane].buffer_type     = (props->GetFormat() == XMA_VPE_FMT_TYPE) ? XMA_DEVICE_ONLY_BUFFER_TYPE : XMA_HOST_BUFFER_TYPE;
      mem                                  = static_cast<VpiLinearMem*>(vpi_memory_calloc(1, sizeof(VpiLinearMem)));
      if (!mem) {
        for (--plane; plane >= 0; --plane) {
          vpi_memory_free(m_frame->data[plane].buffer);
        }
        THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::InternalFrame() Out Of Memory!");
      }
      m_frame->data[plane].buffer   = mem;
      m_frame->data[plane].is_clone = false;
    }
    for (; plane < XMA_MAX_PLANES; ++plane) {
      m_frame->frame_props.linesize[plane] = 0;
      m_frame->data[plane].buffer_type     = NO_BUFFER;
      m_frame->data[plane].buffer          = nullptr;
      m_frame->data[plane].host            = nullptr;
      m_frame->data[plane].device          = nullptr;
    }
    mem0                    = buffer->m_handle;
    mem                     = static_cast<VpiLinearMem*>(m_frame->data[0].buffer);
    mem->size               = sizes[0]; // cppcheck-suppress [uninitvar]
    mem->virtual_address_rc = mem0->virtual_address_rc;
    mem->bus_address_ep     = mem0->bus_address_ep;
    mem->task_id            = mem0->task_id;
    mem->mem_type           = mem0->mem_type;
    m_frame->data[0].host   = static_cast<uint8_t*>(mem->virtual_address_rc);
    m_frame->data[0].device = reinterpret_cast<void*>(mem->bus_address_ep);
    for (plane = 1; plane < planes; ++plane) {
      mem0                        = static_cast<VpiLinearMem*>(m_frame->data[plane - 1].buffer);
      mem                         = static_cast<VpiLinearMem*>(m_frame->data[plane].buffer);
      mem->size                   = sizes[plane];
      mem->bus_address_ep         = mem0->bus_address_ep + sizes[plane - 1];
      mem->virtual_address_rc     = static_cast<uint8_t*>(mem0->virtual_address_rc) + sizes[plane - 1];
      mem->task_id                = mem0->task_id;
      mem->mem_type               = mem0->mem_type;
      m_frame->data[plane].host   = static_cast<uint8_t*>(mem->virtual_address_rc);
      m_frame->data[plane].device = reinterpret_cast<void*>(mem->bus_address_ep);
    }
  }

  auto info                     = std::make_unique<PrivateFrameInfo>(sf, m_log, std::move(buffer));
  m_frame->private_do_not_touch = static_cast<void*>(info.release());

  IncRef();

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  m_sf->m_frames.push_back(m_frame.get());
  lock.unlock();
#endif
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalFrame::InternalFrame(IntSessionFactoryPtr sf, LogPtr log, FramePropertiesPtr props, uint8_t** frame_data, std::function<void(uint8_t**)> free_callback,
    XmaFraction frame_rate, int64_t pts, bool is_idr) :
    m_props(props),
    m_sf(sf), m_log(log) {
  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame(format=%s, sw_format=%s, size=%dx%d)", InternalFrameProperties::GetPixFmtName(props->GetFormat()),
      InternalFrameProperties::GetPixFmtName(props->GetSwFormat()), props->GetWidth(), props->GetHeight());
  m_frame = std::make_unique<XmaFrame>();
  memset(m_frame.get(), 0, sizeof(XmaFrame));
  m_frame->frame_props.format    = props->GetFormat();
  m_frame->frame_props.sw_format = props->GetSwFormat();
  m_frame->frame_props.width     = props->GetWidth();
  m_frame->frame_props.height    = props->GetHeight();
  m_frame->frame_props.flags     = props->GetFlags();
  m_frame->frame_rate            = frame_rate;
  m_frame->pts                   = pts;
  m_frame->is_idr                = is_idr;

  auto info = std::make_unique<PrivateFrameInfo>(sf, log);
  info->SetCallback(free_callback);

  int32_t planes = props->GetNumPlanes();
  int32_t plane;
  for (plane = 0; plane < planes; ++plane) {
    m_frame->frame_props.linesize[plane] = props->GetPlaneStride(plane);
    m_frame->data[plane].buffer_type     = XMA_HOST_BUFFER_TYPE;
    m_frame->data[plane].buffer          = nullptr;
    m_frame->data[plane].host            = frame_data[plane];
    m_frame->data[plane].device          = nullptr;
    m_frame->data[plane].is_clone        = true;
  }
  for (; plane < XMA_MAX_PLANES; ++plane) {
    m_frame->frame_props.linesize[plane] = 0;
    m_frame->data[plane].buffer_type     = NO_BUFFER;
    m_frame->data[plane].buffer          = nullptr;
    m_frame->data[plane].host            = nullptr;
    m_frame->data[plane].device          = nullptr;
  }

  m_frame->private_do_not_touch = static_cast<void*>(info.release());

  IncRef();

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  m_sf->m_frames.push_back(m_frame.get());
  lock.unlock();
#endif
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalFrame::InternalFrame(IntSessionFactoryPtr sf, LogPtr log, FramePtr frame, XmaFraction frame_rate, int64_t pts, bool is_idr) :
    m_props(frame->GetProps()), m_sf(sf), m_log(log) {
  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame(format=%s, sw_format=%s, size=%dx%d)", InternalFrameProperties::GetPixFmtName(m_props->GetFormat()),
      InternalFrameProperties::GetPixFmtName(m_props->GetSwFormat()), m_props->GetWidth(), m_props->GetHeight());
  m_frame = std::make_unique<XmaFrame>();
  memset(m_frame.get(), 0, sizeof(XmaFrame));
  m_frame->frame_props.format    = m_props->GetFormat();
  m_frame->frame_props.sw_format = m_props->GetSwFormat();
  m_frame->frame_props.width     = m_props->GetWidth();
  m_frame->frame_props.height    = m_props->GetHeight();
  m_frame->frame_props.flags     = m_props->GetFlags();
  m_frame->frame_rate            = frame_rate;
  m_frame->pts                   = pts;
  m_frame->is_idr                = is_idr;

  auto info = std::make_unique<PrivateFrameInfo>(sf, log);
  info->SetParent(frame);

  int32_t planes = m_props->GetNumPlanes();
  int32_t plane;
  for (plane = 0; plane < planes; ++plane) {
    m_frame->frame_props.linesize[plane] = m_props->GetPlaneStride(plane);
    m_frame->data[plane].buffer_type     = frame->GetBufferType();
    m_frame->data[plane].buffer          = frame->GetBuffer(plane);
    m_frame->data[plane].host            = frame->GetHostBuffer(plane);
    m_frame->data[plane].device          = frame->GetDeviceBuffer(plane);
    m_frame->data[plane].is_clone        = true;
  }
  for (; plane < XMA_MAX_PLANES; ++plane) {
    m_frame->frame_props.linesize[plane] = 0;
    m_frame->data[plane].buffer_type     = NO_BUFFER;
    m_frame->data[plane].buffer          = nullptr;
    m_frame->data[plane].host            = nullptr;
    m_frame->data[plane].device          = nullptr;
  }

  m_frame->private_do_not_touch = static_cast<void*>(info.release());

  IncRef();

  CopySideData(frame);

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  m_sf->m_frames.push_back(m_frame.get());
  lock.unlock();
#endif
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalFrame::InternalFrame(
    IntSessionFactoryPtr sf, LogPtr log, VpiFrm* pic, XmaFraction frame_rate, XmaFormatType format, XmaFormatType sw_format, int32_t width, int32_t height) :
    m_sf(sf),
    m_log(log) {
  if (!pic->valid) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::InternalFrame() Invalid pic!");
  }

  if ((width == 0) || (height == 0)) {
    width  = pic->width;
    height = pic->height;
  }

  int32_t ch;
  for (ch = 0; ch < VPE_CHANNEL_NUM; ++ch) {
    if (pic->ch_info[ch].enabled && (pic->ch_info[ch].width == width) && (pic->ch_info[ch].height == height) && (pic->ch_info[ch].format == pic->format)) {
      break;
    }
  }
  if (ch == VPE_CHANNEL_NUM) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::InternalFrame() Unable to find channel!");
  }

  if ((format == XMA_NONE_FMT_TYPE) || (sw_format == XMA_NONE_FMT_TYPE)) {
    format    = static_cast<XmaFormatType>(pic->ch_info[ch].format);
    sw_format = static_cast<XmaFormatType>(pic->ch_info[ch].sw_format);
  } else {
    for (; ch < VPE_CHANNEL_NUM; ++ch) {
      if (pic->ch_info[ch].enabled && (static_cast<XmaFormatType>(pic->ch_info[ch].format) == format) &&
          (static_cast<XmaFormatType>(pic->ch_info[ch].sw_format) == sw_format) && (pic->ch_info[ch].width == width) && (pic->ch_info[ch].height == height) &&
          (pic->ch_info[ch].format == pic->format)) {
        break;
      }
    }
    if (ch == VPE_CHANNEL_NUM) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::InternalFrame() Unable to find channel!");
    }
  }
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame(format=%s, sw_format=%s, size=%dx%d)", InternalFrameProperties::GetPixFmtName(format),
      InternalFrameProperties::GetPixFmtName(sw_format), width, height);

  int32_t planes = vpi_frame_get_planes(sw_format);

  VpiBuffer* vpi_buf = static_cast<VpiBuffer*>(pic->vpi_picinfo_buf);
  if (vpi_buf) {
    PrivateVpiFrmInfo* info = static_cast<PrivateVpiFrmInfo*>(pic->opaque);
    if (info && info->IsValid()) {
      vpi_frame_info_t vpi_info = info->GetFrame(format, sw_format, width, height);
      if (vpi_info) {
        m_frame = std::make_unique<XmaFrame>();
        memset(m_frame.get(), 0, sizeof(XmaFrame));

        m_frame->frame_props.format    = format;
        m_frame->frame_props.sw_format = sw_format;
        m_frame->frame_props.width     = width;
        m_frame->frame_props.height    = height;
        m_frame->frame_props.flags     = pic->ch_info[ch].flag;
        m_frame->frame_rate            = frame_rate;
        m_frame->pts                   = pic->pts;
        m_frame->is_idr                = pic->key_frame;

        int32_t plane;
        for (plane = 0; plane < XMA_MAX_PLANES; ++plane) {
          m_frame->frame_props.linesize[plane] = 0;
          m_frame->data[plane].buffer_type     = NO_BUFFER;
          m_frame->data[plane].buffer          = nullptr;
          m_frame->data[plane].host            = nullptr;
          m_frame->data[plane].device          = nullptr;
        }

        CopyFrame(m_log, m_frame.get(), std::move(vpi_info->frame));
        CopySideData(pic);
        m_props = std::make_shared<InternalFrameProperties>(m_sf, m_log, &m_frame->frame_props);
        if (info->GetFrameCount() == 0) {
          vpi_buf->freebuf = info->GetFreeBuf();
          pic->opaque      = nullptr;
          delete info;
        }

#ifdef XMA_CHECK_MEM
        std::unique_lock<std::mutex> lock(m_sf->m_mutex);
        m_sf->m_frames.push_back(m_frame.get());
        lock.unlock();
#endif
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame() - Returning");
        return;
      }
    }
  }

  m_frame   = std::make_unique<XmaFrame>();
  auto info = std::make_unique<PrivateFrameInfo>(sf, log);

  memset(m_frame.get(), 0, sizeof(XmaFrame));

  XmaBufferType buffer_type;
  switch (vpi_buf->m_handle->mem_type & (RC_SIDE_EN | EP_SIDE_EN)) {
  case RC_SIDE_EN:
    buffer_type = XMA_HOST_BUFFER_TYPE;
    break;
  case EP_SIDE_EN:
    buffer_type = XMA_DEVICE_ONLY_BUFFER_TYPE;
    break;
  case (RC_SIDE_EN | EP_SIDE_EN):
    buffer_type = XMA_DEVICE_BUFFER_TYPE;
    break;
  case 0:
    buffer_type = NO_BUFFER;
    break;
  }

  int32_t plane;
  for (plane = 0; plane < planes; ++plane) {
    VpiLinearMem* mem                    = static_cast<VpiLinearMem*>(vpi_memory_calloc(1, sizeof(VpiLinearMem)));
    mem->size                            = vpi_frame_get_plane_size(pic->ch_info[ch].sw_format, plane, width, pic->ch_info[ch].walign, height, pic->ch_info[ch].halign);
    mem->request_size                    = mem->size;
    mem->virtual_address_rc              = vpi_buf->buf_vaddr[ch][plane];
    mem->bus_address_ep                  = reinterpret_cast<addr_t>(vpi_buf->buf_paddr[ch][plane]);
    m_frame->data[plane].buffer          = mem;
    m_frame->data[plane].buffer_type     = buffer_type;
    m_frame->data[plane].host            = static_cast<uint8_t*>(mem->virtual_address_rc);
    m_frame->data[plane].device          = reinterpret_cast<void*>(mem->bus_address_ep);
    m_frame->data[plane].is_clone        = false;
    m_frame->frame_props.linesize[plane] = pic->ch_info[ch].stride_size[plane];
  }
  for (; plane < XMA_MAX_PLANES; ++plane) {
    m_frame->frame_props.linesize[plane] = 0;
    m_frame->data[plane].buffer_type     = NO_BUFFER;
    m_frame->data[plane].buffer          = nullptr;
    m_frame->data[plane].host            = nullptr;
    m_frame->data[plane].device          = nullptr;
  }

  m_frame->frame_props.format    = static_cast<XmaFormatType>(pic->ch_info[ch].format);
  m_frame->frame_props.sw_format = static_cast<XmaFormatType>(pic->ch_info[ch].sw_format);
  m_frame->frame_props.width     = width;
  m_frame->frame_props.height    = height;
  m_frame->frame_props.flags     = pic->ch_info[ch].flag;
  m_frame->frame_rate            = frame_rate;
  m_frame->pts                   = pic->pts;
  m_frame->is_idr                = pic->key_frame;
  info->AddVpiFrm(pic);

  m_props = std::make_shared<InternalFrameProperties>(m_sf, m_log, &m_frame->frame_props);

  m_frame->private_do_not_touch = static_cast<void*>(info.release());

  IncRef();
  CopySideData(pic);

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  m_sf->m_frames.push_back(m_frame.get());
  lock.unlock();
#endif
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalFrame::~InternalFrame() {
  if (!m_frame) {
    return;
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::~InternalFrame(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(m_frame->frame_props.format), InternalFrameProperties::GetPixFmtName(m_frame->frame_props.sw_format),
      m_frame->frame_props.width, m_frame->frame_props.height);

  DecRef();

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::~InternalFrame() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
XmaFrame* InternalFrame::GetFrame() {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetFrame", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetFrame(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(m_frame->frame_props.format), InternalFrameProperties::GetPixFmtName(m_frame->frame_props.sw_format),
      m_frame->frame_props.width, m_frame->frame_props.height);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetFrame() - Returning");

  return m_frame.get();
}

///////////////////////////////////////////////////////////////////////////////
VpiBufferPool* InternalFrame::GetBufferPool() {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetBufferPool", XMA_FRAME, m_frame);

  PrivateFrameInfo* info = static_cast<PrivateFrameInfo*>(m_frame->private_do_not_touch);
  if ((info) && (info->IsValid())) {
    return info->GetBufferPool();
  }

  return nullptr;
}

///////////////////////////////////////////////////////////////////////////////
void InternalFrame::ExtendBufferPool(uint32_t frames) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::ExtendBufferPool", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::ExtendBufferPool(format=%s, sw_format=%s, size=%dx%d, extend=%u)",
      InternalFrameProperties::GetPixFmtName(m_frame->frame_props.format), InternalFrameProperties::GetPixFmtName(m_frame->frame_props.sw_format),
      m_frame->frame_props.width, m_frame->frame_props.height, frames);

  VpiBufferPool*    pool = nullptr;
  PrivateFrameInfo* info = static_cast<PrivateFrameInfo*>(m_frame->private_do_not_touch);
  if ((info) && (info->IsValid())) {
    pool = info->GetBufferPool();
  }

  VpiRet ret = VPI_ERR_SW;
  if (pool) {
    int32_t max_capacity = vpi_buffer_pool_get_capacity(pool);
    int32_t curr_depth   = vpi_buffer_pool_get_depth(pool);
    int32_t extend_num   = frames;
    int32_t new_depth    = curr_depth + extend_num;
    if (max_capacity) {
      if (new_depth > max_capacity) {
        extend_num = max_capacity - curr_depth;
      }
    }
    ret = vpi_buffer_pool_extend(pool, extend_num);
  }
  if (ret == VPI_SUCCESS) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::ExtendBufferPool() - returned success");
  } else {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::ExtendBufferPool() - returned");
  }
}

///////////////////////////////////////////////////////////////////////////////
void InternalFrame::SetSelf() {
  LOCK_XMA_FRAME(m_log, "InternalFrame::SetSelf", XMA_FRAME, m_frame);

  PrivateFrameInfo* info = static_cast<PrivateFrameInfo*>(m_frame->private_do_not_touch);
  if ((info) && (info->IsValid())) {
    info->SetSelf(std::static_pointer_cast<InternalFrame>(shared_from_this()));
  }
}

///////////////////////////////////////////////////////////////////////////////
void InternalFrame::SetFrozen(bool sidedata_only) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::SetFrozen", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::SetFrozen(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(m_frame->frame_props.format), InternalFrameProperties::GetPixFmtName(m_frame->frame_props.sw_format),
      m_frame->frame_props.width, m_frame->frame_props.height);

  if (m_frame->private_do_not_touch && ((!sidedata_only) || (m_sf->GetAPIVersion() >= XMA_API_VERSION_1_1))) {
    PrivateFrameInfo* info = static_cast<PrivateFrameInfo*>(m_frame->private_do_not_touch);
    if (info->IsValid()) {
      info->SetFrozen();
    }
  }

  if (m_frame->side_data) {
    for (uint32_t index = 0; m_frame->side_data[index]; ++index) {
      InternalSideData::GetIntSideData(m_log, m_frame->side_data[index])->SetFrozen();
    }
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::SetFrozen() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
bool InternalFrame::IsFrozen() const {
  LOCK_XMA_FRAME(m_log, "InternalFrame::IsFrozen", XMA_FRAME, m_frame);

  if (m_frame->private_do_not_touch) {
    PrivateFrameInfo* info = static_cast<PrivateFrameInfo*>(m_frame->private_do_not_touch);
    if (info->IsValid()) {
      return info->IsFrozen();
    }
  }

  THROW_RUNTIME_EXCEPTION(
      m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::IsFrozen() XmaFrame was not created by xma_frame_alloc(), xma_frame_from_buffers_clone() or xma_frame_clone()!");
}

///////////////////////////////////////////////////////////////////////////////
void InternalFrame::CopySideData(FramePtr src) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::CopySideData", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::CopySideData(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(m_frame->frame_props.format), InternalFrameProperties::GetPixFmtName(m_frame->frame_props.sw_format),
      m_frame->frame_props.width, m_frame->frame_props.height);

  IntFramePtr int_src = std::dynamic_pointer_cast<InternalFrame>(src);
  if (!int_src) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopySideData() Invalid source!");
  }

  if (m_frame->side_data) {
    ClearAllSideData();
  }

  if (!int_src->m_frame->side_data) {
    return;
  }

  XmaFrameSideData** src_sidedata_array = int_src->m_frame->side_data;
  uint32_t           count;
  for (count = 0; src_sidedata_array[count]; ++count)
    ;

  if (count > 0) {
    XmaFrameSideData** sidedata_array = new XmaFrameSideData*[count + 1];
    if (!sidedata_array) {
      THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopySideData() Out of memory!");
    }
    memset(sidedata_array, 0, sizeof(XmaFrameSideData*) * (count + 1));
    m_frame->side_data = sidedata_array;

    for (uint32_t index = 0; index < count; ++index) {
      m_frame->side_data[index] = src_sidedata_array[index];
      IntSideDataPtr sd_ptr     = InternalSideData::GetIntSideData(m_log, src_sidedata_array[index]);
      sd_ptr->IncRef(true);
    }
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::CopySideData() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
void InternalFrame::CopyUserSideData(FramePtr src) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::CopyUserSideData", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::CopyUserSideData(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(m_frame->frame_props.format), InternalFrameProperties::GetPixFmtName(m_frame->frame_props.sw_format),
      m_frame->frame_props.width, m_frame->frame_props.height);

  if (m_sf->GetAPIVersion() < XMA_API_VERSION_1_1) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopyUserSideData() Unsupported feature in this version of the API!");
  }

  IntFramePtr int_src = std::dynamic_pointer_cast<InternalFrame>(src);
  if (!int_src) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopyUserSideData() Invalid source!");
  }

  uint32_t           user_count = 0;
  uint32_t           index;
  XmaFrameSideData*  sd;
  XmaFrameSideData** src_sidedata_array = int_src->m_frame->side_data;
  if (src_sidedata_array) {
    for (index = 0; (sd = src_sidedata_array[index]); ++index) {
      if (sd->type >= XMA_FRAME_SIDE_DATA_USER_DEFINED_START) {
        user_count++;
      }
    }
  }

  uint32_t           count              = 0;
  XmaFrameSideData** dst_sidedata_array = m_frame->side_data;
  if (dst_sidedata_array) {
    for (; (sd = dst_sidedata_array[count]); ++count)
      ;
  }

  uint32_t           new_count          = user_count + count + 1;
  XmaFrameSideData** new_sidedata_array = new XmaFrameSideData*[new_count];
  if (!new_sidedata_array) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopyUserSideData() Out of memory!");
  }
  memset(new_sidedata_array, 0, sizeof(XmaFrameSideData*) * new_count);

  if (dst_sidedata_array) {
    for (index = 0; (sd = dst_sidedata_array[index]); ++index) {
      new_sidedata_array[index] = dst_sidedata_array[index];
    }
    delete[] dst_sidedata_array;
  }
  if (src_sidedata_array) {
    for (index = 0; (sd = src_sidedata_array[index]); ++index) {
      if (sd->type >= XMA_FRAME_SIDE_DATA_USER_DEFINED_START) {
        new_sidedata_array[count++] = sd;
        IntSideDataPtr sd_ptr       = InternalSideData::GetIntSideData(m_log, sd);
        sd_ptr->IncRef(true);
      }
    }
  }

  m_frame->side_data = new_sidedata_array;

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::CopyUserSideData() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
void InternalFrame::CopySideDataMetadata(FramePtr src) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::CopySideDataMetadata", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::CopySideDataMetadata(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(m_frame->frame_props.format), InternalFrameProperties::GetPixFmtName(m_frame->frame_props.sw_format),
      m_frame->frame_props.width, m_frame->frame_props.height);

  if (m_sf->GetAPIVersion() < XMA_API_VERSION_1_1) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopySideDataMetadata() Unsupported feature in this version of the API!");
  }

  IntFramePtr int_src = std::dynamic_pointer_cast<InternalFrame>(src);
  if (!int_src) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopySideDataMetadata() Invalid source!");
  }

  if ((!int_src->m_frame->side_data) || (!m_frame->side_data)) {
    return;
  }

  XmaFrameSideData** src_sidedata_array = int_src->m_frame->side_data;
  XmaFrameSideData** dst_sidedata_array = m_frame->side_data;
  XmaFrameSideData*  src_sidedata;
  for (uint32_t src_index = 0; (src_sidedata = src_sidedata_array[src_index]); ++src_index) {
    IntSideDataPtr    src_sd_ptr = InternalSideData::GetIntSideData(m_log, src_sidedata);
    XmaFrameSideData* dst_sidedata;
    for (uint32_t dst_index = 0; (dst_sidedata = dst_sidedata_array[dst_index]); ++dst_index) {
      if ((src_sidedata->buffer_type == dst_sidedata->buffer_type) && (src_sidedata->host == dst_sidedata->host) && (src_sidedata->device == dst_sidedata->device) &&
          (src_sidedata->size == dst_sidedata->size)) {
        IntSideDataPtr dst_sd_ptr = InternalSideData::GetIntSideData(m_log, dst_sidedata);
        dst_sd_ptr->CopyMetadata(src_sd_ptr);
      }
    }
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::CopySideDataMetadata() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
SideDataPtr InternalFrame::GetSideDataAtIndex(uint32_t index) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetSideDataAtIndex", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetSideDataAtIndex(frame=(format=%s, sw_format=%s, size=%dx%d), index=%u)",
      InternalFrameProperties::GetPixFmtName(GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(GetProps()->GetSwFormat()), GetProps()->GetWidth(),
      GetProps()->GetHeight(), index);

  if (!m_frame->side_data) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetSideDataAtIndex() - Returning side data not found");
    return nullptr;
  }

  XmaFrameSideData** sidedata_array = m_frame->side_data;

  if (!sidedata_array[index]) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetSideDataAtIndex() - Returning side data not found");
    return nullptr;
  }

  IntSideDataPtr ret = InternalSideData::GetIntSideData(m_log, sidedata_array[index]);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetSideDataAtIndex() - Returning side data");
  return ret;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrame::IncRef() {
  LOCK_XMA_FRAME(m_log, "InternalFrame::IncRef", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::IncRef(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(m_frame->frame_props.format), InternalFrameProperties::GetPixFmtName(m_frame->frame_props.sw_format),
      m_frame->frame_props.width, m_frame->frame_props.height);

  int32_t planes = m_props->GetNumPlanes();
  for (int32_t plane = 0; plane < planes; ++plane) {
    ++m_frame->data[plane].refcount;
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::IncRef() - Returning");
  return m_frame->data[0].refcount;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrame::DecRef() {
  LOCK_XMA_FRAME(m_log, "InternalFrame::DecRef", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::DecRef(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(m_frame->frame_props.format), InternalFrameProperties::GetPixFmtName(m_frame->frame_props.sw_format),
      m_frame->frame_props.width, m_frame->frame_props.height);

  int32_t planes = m_props->GetNumPlanes();
  for (int32_t plane = 0; plane < planes; ++plane) {
    --m_frame->data[plane].refcount;
  }

  if (m_frame->data[0].refcount > 0) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::DecRef() - Returning");
    return m_frame->data[0].refcount;
  }

  if (m_frame->private_do_not_touch) {
    PrivateFrameInfo* info = static_cast<PrivateFrameInfo*>(m_frame->private_do_not_touch);
    if (info->IsValid()) {
      info->Unfreeze();
      ClearAllSideData();
      if (info->RemoveVpiFrm() == XMA_SUCCESS) { // if the frame was made from a VpiFrm...

        m_frame->private_do_not_touch = nullptr;
        for (int32_t plane = 0; plane < planes; ++plane) {
          vpi_memory_free(m_frame->data[plane].buffer);
        }
        memset(m_frame->data, 0, sizeof(XmaBufferRef) * XMA_MAX_PLANES);

#ifdef XMA_CHECK_MEM
        std::unique_lock<std::mutex> lock(m_sf->m_mutex);
        if (m_sf->m_frames.remove(m_frame.get()) != 1) {
          m_log->LogMsg(XMA_ERROR_LOG, XMA_FRAME, "XMA Memory check Error! Frame %p removed twice?", m_frame.get());
        }
        lock.unlock();
#endif

        guard.unlock();
        delete info;
        m_frame = nullptr;

        m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::DecRef() - Returning");
        return 0;
      }
      if (m_frame->data[0].is_clone) {
        uint8_t* data[XMA_MAX_PLANES];
        memset(&data, 0, sizeof(data));
        for (int32_t plane = 0; plane < planes; ++plane) {
          data[plane] = m_frame->data[plane].host;
        }
        info->CallCallback(data);
      }
      guard.unlock();
      delete info;
      m_frame->private_do_not_touch = nullptr;
    }
  }

  switch (m_frame->data[0].buffer_type) {
  case XMA_HOST_BUFFER_TYPE:
  case XMA_DEVICE_ONLY_BUFFER_TYPE:
  case XMA_DEVICE_BUFFER_TYPE:
    for (int32_t plane = 0; plane < planes; ++plane) {
      if ((m_frame->data[plane].buffer) && (!m_frame->data[plane].is_clone)) {
        vpi_memory_free(static_cast<void*>(m_frame->data[plane].buffer));
      }

      m_frame->data[plane].buffer      = nullptr;
      m_frame->data[plane].host        = nullptr;
      m_frame->data[plane].device      = nullptr;
      m_frame->data[plane].buffer_type = NO_BUFFER;
    }
    break;
  case NO_BUFFER:
    break;
  default:
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::DecRef() Invalid buffer type!");
  }

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  if (m_sf->m_frames.remove(m_frame.get()) != 1) {
    m_log->LogMsg(XMA_ERROR_LOG, XMA_FRAME, "XMA Memory check Error! Frame %p removed twice?", m_frame.get());
  }
  lock.unlock();
#endif

  m_frame = nullptr;

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::DecRef() - Returning");
  return 0;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrame::GetRefCount() {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetRefCount", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetRefCount(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(m_frame->frame_props.format), InternalFrameProperties::GetPixFmtName(m_frame->frame_props.sw_format),
      m_frame->frame_props.width, m_frame->frame_props.height);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetRefCount() - Returning");
  return m_frame->data[0].refcount;
}

///////////////////////////////////////////////////////////////////////////////
XmaBufferType InternalFrame::GetBufferType() const {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetBufferType", XMA_FRAME, m_frame);

  return m_frame->data[0].buffer_type;
}

///////////////////////////////////////////////////////////////////////////////
void* InternalFrame::GetBuffer(int32_t plane) const {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetBuffer", XMA_FRAME, m_frame);

  if ((plane < 0) || (plane >= m_props->GetNumPlanes())) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetBuffer() Invalid parameter!");
  }

  return m_frame->data[plane].buffer;
}

///////////////////////////////////////////////////////////////////////////////
uint8_t* InternalFrame::GetHostBuffer(int32_t plane) const {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetHostBuffer", XMA_FRAME, m_frame);

  if ((plane < 0) || (plane >= m_props->GetNumPlanes())) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetHostBuffer() Invalid parameter!");
  }

  switch (m_frame->data[0].buffer_type) {
  case NO_BUFFER:
  case XMA_DEVICE_ONLY_BUFFER_TYPE:
    return nullptr;
  case XMA_HOST_BUFFER_TYPE:
  case XMA_DEVICE_BUFFER_TYPE:
    return m_frame->data[plane].host;
  default:
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetHostBuffer() Invalid buffer type!");
  }
}

///////////////////////////////////////////////////////////////////////////////
void* InternalFrame::GetDeviceBuffer(int32_t plane) const {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetDeviceBuffer", XMA_FRAME, m_frame);

  if ((plane < 0) || (plane >= m_props->GetNumPlanes())) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetDeviceBuffer() Invalid parameter!");
  }

  switch (m_frame->data[0].buffer_type) {
  case NO_BUFFER:
  case XMA_HOST_BUFFER_TYPE:
    return nullptr;
  case XMA_DEVICE_ONLY_BUFFER_TYPE:
  case XMA_DEVICE_BUFFER_TYPE:
    return m_frame->data[plane].device;
  default:
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetDeviceBuffer() Invalid buffer type!");
  }
}

///////////////////////////////////////////////////////////////////////////////
FramePropertiesPtr InternalFrame::GetProps() const {
  return m_props;
}

///////////////////////////////////////////////////////////////////////////////
XmaFraction InternalFrame::GetFrameRate() const {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetFrameRate", XMA_FRAME, m_frame);

  return m_frame->frame_rate;
}

///////////////////////////////////////////////////////////////////////////////
int64_t InternalFrame::GetPts() const {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetPts", XMA_FRAME, m_frame);

  return m_frame->pts;
}

///////////////////////////////////////////////////////////////////////////////
int InternalFrame::IsIdr() const {
  LOCK_XMA_FRAME(m_log, "InternalFrame::IsIdr", XMA_FRAME, m_frame);

  return m_frame->is_idr;
}

///////////////////////////////////////////////////////////////////////////////
void InternalFrame::SetFrameRate(XmaFraction frame_rate) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::SetFrameRate", XMA_FRAME, m_frame);

  if (IsFrozen()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::SetFrameRate() Unable to set framerate when frozen!");
  }

  if (frame_rate.denominator <= 0 || frame_rate.numerator <= 0) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::SetFrameRate() Invalid frame_rate!");
  }

  m_frame->frame_rate = frame_rate;
}

///////////////////////////////////////////////////////////////////////////////
void InternalFrame::SetPts(int64_t pts) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::SetPts", XMA_FRAME, m_frame);

  if (IsFrozen()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::SetPts() Unable to set PTS when frozen!");
  }

  m_frame->pts = pts;
}

///////////////////////////////////////////////////////////////////////////////
void InternalFrame::SetIdr(bool is_idr) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::SetIdr", XMA_FRAME, m_frame);

  if (IsFrozen()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::SetIdr() Unable to set IDR when frozen!");
  }

  if (is_idr != false && is_idr != true) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::SetIdr() Invalid is_idr!");
  }

  m_frame->is_idr = is_idr;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrame::AddSideData(SideDataPtr side_data) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::AddSideData", XMA_FRAME, m_frame);

  if (!side_data) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::AddSideData() Invalid parameter!");
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::AddSideData(frame=(format=%s, sw_format=%s, size=%dx%d), side data=(type=%u))",
      InternalFrameProperties::GetPixFmtName(GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(GetProps()->GetSwFormat()), GetProps()->GetWidth(),
      GetProps()->GetHeight(), static_cast<uint32_t>(side_data->GetType()));

  if (IsFrozen()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::AddSideData() Unable to add sidedata when frozen!");
  }

  IntSideDataPtr sd = std::dynamic_pointer_cast<InternalSideData>(side_data);
  if (!sd) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::AddSideData() side data was not created by SessionFactory::AllocSideData()!");
  }

  if (!m_frame->side_data) {
    m_frame->side_data = new XmaFrameSideData*[2];
    if (!m_frame->side_data) {
      THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::AddSideData() Out of memory!");
    }
    m_frame->side_data[0] = sd->GetSideData();
    m_frame->side_data[1] = nullptr;
    sd->IncRef(true);

    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::AddSideData() - Returning");
    return XMA_SUCCESS;
  }

  side_data_buffer_t   vpi_sd = sd->GetVpiSideData();
  XmaFrameSideData*    itr;
  int32_t              index      = -1;
  XmaFrameSideDataType type       = side_data->GetType();
  uint32_t             type_count = 0;
  while ((itr = m_frame->side_data[++index]) != NULL) {
    if (type == itr->type) {
      ++type_count;
    }
    if (itr->private_do_not_touch) {
      PrivateSideDataInfo* info = InternalSideData::GetPrivateInfo(m_log, itr);
      if (info->GetVpiSideData().get() == vpi_sd.get()) {
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::AddSideData() - Returning");
        return XMA_SUCCESS;
      }
    }
  }
  if ((m_sf->GetAPIVersion() < XMA_API_VERSION_1_1) && (type_count > 0)) {
    RemoveSideDataType(type);
    index -= type_count;
  }

  XmaFrameSideData** sidedata_array = new XmaFrameSideData*[index + 2];
  if (!sidedata_array) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::AddSideData() Out of memory!");
  }
  memset(sidedata_array, 0, sizeof(XmaFrameSideData*) * (index + 2));
  index = -1;
  while ((itr = m_frame->side_data[++index]) != nullptr) {
    sidedata_array[index] = itr;
  }
  sidedata_array[index++] = sd->GetSideData();
  sidedata_array[index]   = nullptr;
  sd->IncRef(true);

  delete[] m_frame->side_data;
  m_frame->side_data = sidedata_array;

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::AddSideData() - Returning");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
SideDataPtr InternalFrame::GetSideData(enum XmaFrameSideDataType type) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetSideData", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetSideData(frame=(format=%s, sw_format=%s, size=%dx%d), side data=(type=%u))",
      InternalFrameProperties::GetPixFmtName(GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(GetProps()->GetSwFormat()), GetProps()->GetWidth(),
      GetProps()->GetHeight(), static_cast<uint32_t>(type));

  if (!m_frame->side_data) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetSideData() - Returning side data not found");
    return nullptr;
  }

  XmaFrameSideData** sidedata_array = m_frame->side_data;
  if (!sidedata_array) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetSideData() - Returning side data not found");
    return nullptr;
  }

  uint32_t i;
  for (i = 0; sidedata_array[i]; ++i) {
    if (sidedata_array[i]->type == type) {
      break;
    }
  }

  if (!sidedata_array[i]) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetSideData() - Returning side data not found");
    return nullptr;
  }

  IntSideDataPtr ret = InternalSideData::GetIntSideData(m_log, sidedata_array[i]);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetSideData() - Returning side data");
  return ret;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrame::RemoveSideData(SideDataPtr side_data) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::RemoveSideData", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::RemoveSideData()");

  if (!side_data) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::RemoveSideData() Invalid parameter!");
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::RemoveSideData(frame=(format=%s, sw_format=%s, size=%dx%d), side data=(type=%u))",
      InternalFrameProperties::GetPixFmtName(GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(GetProps()->GetSwFormat()), GetProps()->GetWidth(),
      GetProps()->GetHeight(), static_cast<uint32_t>(side_data->GetType()));

  if (IsFrozen()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::RemoveSideData() Unable to remove sidedata when frozen!");
  }

  if (!m_frame->side_data) {
    THROW_INVALID_ARGUMENT_EXCEPTION(
        m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::RemoveSideData() Frame %p has no side data buffer %p!", m_frame.get(), side_data.get());
  }

  XmaFrameSideData** sidedata_array = m_frame->side_data;

  IntSideDataPtr sd = std::dynamic_pointer_cast<InternalSideData>(side_data);
  if (!sd) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::RemoveSideData() side data was not created by SessionFactory::AllocSideData()!");
  }

  uint32_t i;
  for (i = 0; sidedata_array[i]; ++i) {
    PrivateSideDataInfo* sd_info = InternalSideData::GetPrivateInfo(m_log, sidedata_array[i]);
    if (sd->GetVpiSideData().get() == sd_info->GetVpiSideData().get()) {
      break;
    }
  }

  if (!sidedata_array[i]) {
    THROW_INVALID_ARGUMENT_EXCEPTION(
        m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::RemoveSideData() Frame %p has no side data buffer %p!", m_frame.get(), side_data.get());
  }

  do {
    sidedata_array[i] = sidedata_array[i + 1];
    ++i;
  } while (sidedata_array[i]);

  sd->DecRef(true);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::RemoveSideData() - Returning");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrame::RemoveSideDataType(enum XmaFrameSideDataType type) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::RemoveSideDataType", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::RemoveSideDataType(frame=(format=%s, sw_format=%s, size=%dx%d), type=%u)",
      InternalFrameProperties::GetPixFmtName(GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(GetProps()->GetSwFormat()), GetProps()->GetWidth(),
      GetProps()->GetHeight(), static_cast<uint32_t>(type));

  if (IsFrozen()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::RemoveSideData() Unable to remove sidedata when frozen!");
  }

  if (!m_frame->side_data) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::RemoveSideDataType() - Returning");
    return XMA_SUCCESS;
  }

  XmaFrameSideData** sidedata_array = m_frame->side_data;

  uint32_t dst_index = 0;
  uint32_t src_index = 0;
  for (; sidedata_array[src_index];) {
    if (sidedata_array[src_index]->type == type) {
      IntSideDataPtr sd = InternalSideData::GetIntSideData(m_log, sidedata_array[src_index]);
      sd->DecRef(true);
      ++src_index;
    } else {
      sidedata_array[dst_index++] = sidedata_array[src_index++];
    }
  }
  sidedata_array[dst_index] = nullptr;

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::RemoveSideDataType() - Returning");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
void InternalFrame::ClearAllSideData() {
  LOCK_XMA_FRAME(m_log, "InternalFrame::ClearAllSideData", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::ClearAllSideData(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(GetProps()->GetSwFormat()), GetProps()->GetWidth(),
      GetProps()->GetHeight());

  if (IsFrozen()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::ClearAllSideData() Unable to clear all sidedata when frozen!");
  }

  if (!m_frame->side_data) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::ClearAllSideData() - Returning");
    return;
  }

  XmaFrameSideData** sidedata_array = m_frame->side_data;
  for (uint32_t index = 0; sidedata_array[index]; ++index) {
    IntSideDataPtr sd = InternalSideData::GetIntSideData(m_log, sidedata_array[index]);
    sd->DecRef(true);
  }
  delete[] sidedata_array;
  m_frame->side_data = nullptr;

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::ClearAllSideData() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
SideDataPtr InternalFrame::GetFirstSideData() {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetSideData", XMA_FRAME, m_frame);

  if (m_sf->GetAPIVersion() < XMA_API_VERSION_1_1) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetFirstSideData() Unsupported feature in this version of the API!");
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetFirstSideData(frame=(format=%s, sw_format=%s, size=%dx%d))",
      InternalFrameProperties::GetPixFmtName(GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(GetProps()->GetSwFormat()), GetProps()->GetWidth(),
      GetProps()->GetHeight());

  if (!m_frame->side_data) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetFirstSideData() - Returning side data not found");
    return nullptr;
  }

  XmaFrameSideData** sidedata_array = m_frame->side_data;
  if (!sidedata_array) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetFirstSideData() - Returning side data not found");
    return nullptr;
  }

  if (!sidedata_array[0]) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetFirstSideData() - Returning side data not found");
    return nullptr;
  }

  IntSideDataPtr ret = InternalSideData::GetIntSideData(m_log, sidedata_array[0]);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetFirstSideData() - Returning side data");
  return ret;
}

///////////////////////////////////////////////////////////////////////////////
SideDataPtr InternalFrame::GetNextSideData(SideDataPtr side_data) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetNextSideData", XMA_FRAME, m_frame);

  if (m_sf->GetAPIVersion() < XMA_API_VERSION_1_1) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetNextSideData() Unsupported feature in this version of the API!");
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetNextSideData(frame=(format=%s, sw_format=%s, size=%dx%d))",
      InternalFrameProperties::GetPixFmtName(GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(GetProps()->GetSwFormat()), GetProps()->GetWidth(),
      GetProps()->GetHeight());

  if (!side_data) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetNextSideData() Invalid parameter!");
  }

  if (!m_frame->side_data) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetNextSideData() - Returning side data not found");
    return nullptr;
  }

  XmaFrameSideData** sidedata_array = m_frame->side_data;
  if (!sidedata_array) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetNextSideData() - Returning side data not found");
    return nullptr;
  }

  IntSideDataPtr sd = std::dynamic_pointer_cast<InternalSideData>(side_data);
  if (!sd) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetNextSideData() side data was not created by SessionFactory::AllocSideData()!");
  }

  uint32_t i;
  for (i = 0; sidedata_array[i]; ++i) {
    PrivateSideDataInfo* sd_info = InternalSideData::GetPrivateInfo(m_log, sidedata_array[i]);
    if (sd->GetVpiSideData().get() == sd_info->GetVpiSideData().get()) {
      break;
    }
  }

  if (!sidedata_array[i++]) {
    THROW_INVALID_ARGUMENT_EXCEPTION(
        m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetNextSideData() Frame %p has no side data buffer %p!", m_frame.get(), side_data.get());
  }

  if (!sidedata_array[i]) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetNextSideData() - Returning side data not found");
    return nullptr;
  }

  IntSideDataPtr ret = InternalSideData::GetIntSideData(m_log, sidedata_array[i]);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetNextSideData() - Returning side data");
  return ret;
}

///////////////////////////////////////////////////////////////////////////////
SideDataPtr InternalFrame::GetNextSideDataOfSameType(SideDataPtr side_data) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetNextSideDataOfSameType", XMA_FRAME, m_frame);

  if (m_sf->GetAPIVersion() < XMA_API_VERSION_1_1) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetNextSideDataOfSameType() Unsupported feature in this version of the API!");
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetNextSideDataOfSameType(frame=(format=%s, sw_format=%s, size=%dx%d))",
      InternalFrameProperties::GetPixFmtName(GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(GetProps()->GetSwFormat()), GetProps()->GetWidth(),
      GetProps()->GetHeight());

  if (!side_data) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetNextSideDataOfSameType() Invalid parameter!");
  }

  if (!m_frame->side_data) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetNextSideDataOfSameType() - Returning side data not found");
    return nullptr;
  }

  XmaFrameSideData** sidedata_array = m_frame->side_data;
  if (!sidedata_array) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetNextSideDataOfSameType() - Returning side data not found");
    return nullptr;
  }

  IntSideDataPtr sd = std::dynamic_pointer_cast<InternalSideData>(side_data);
  if (!sd) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetNextSideDataOfSameType() side data was not created by SessionFactory::AllocSideData()!");
  }

  uint32_t i;
  for (i = 0; sidedata_array[i]; ++i) {
    PrivateSideDataInfo* sd_info = InternalSideData::GetPrivateInfo(m_log, sidedata_array[i]);
    if (sd->GetVpiSideData().get() == sd_info->GetVpiSideData().get()) {
      break;
    }
  }

  if (!sidedata_array[i++]) {
    THROW_INVALID_ARGUMENT_EXCEPTION(
        m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetNextSideDataOfSameType() Frame %p has no side data buffer %p!", m_frame.get(), side_data.get());
  }

  for (; sidedata_array[i]; ++i) {
    if (sidedata_array[i]->type == side_data->GetType()) {
      break;
    }
  }

  if (!sidedata_array[i]) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetNextSideDataOfSameType() - Returning side data not found");
    return nullptr;
  }

  IntSideDataPtr ret = InternalSideData::GetIntSideData(m_log, sidedata_array[i]);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetNextSideDataOfSameType() - Returning side data");
  return ret;
}

///////////////////////////////////////////////////////////////////////////////
void InternalFrame::CopySideData(VpiFrm* src) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::CopySideData", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::CopySideData(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(GetProps()->GetSwFormat()), GetProps()->GetWidth(),
      GetProps()->GetHeight());

  if (!src) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopySideData() Source Vpi Frame must be specified!");
  }

  XmaAPIVersion api_version        = m_sf->GetAPIVersion();
  bool          has_dyn_enc_params = (src->dynamic_params.is_spatial_aq_gain_changed || src->dynamic_params.is_temporal_aq_gain_changed ||
      src->dynamic_params.is_temporal_mode_changed || src->dynamic_params.is_spatial_mode_changed || src->dynamic_params.is_bit_rate_changed ||
      src->dynamic_params.is_b_frames_changed || src->dynamic_params.is_min_qp_changed || src->dynamic_params.is_max_qp_changed);
  if (api_version >= XMA_API_VERSION_1_1_2) {
    has_dyn_enc_params |= (src->dynamic_params.is_min_bit_rate_changed || src->dynamic_params.is_max_bit_rate_changed || src->dynamic_params.is_qp_changed ||
        src->dynamic_params.is_qp_i_offset_changed || src->dynamic_params.is_qp_b_offset_changed);
  }

  uint32_t count     = (has_dyn_enc_params ? 1 : 0);
  bool     needs_hdr = false;
  if (api_version < XMA_API_VERSION_1_1) {
    if (vpifrm_get_sidedata_count_by_type(src, VPI_SB_VUI_PARAMS) > 0) {
      ++count;
    }
    for (uint32_t type = VPI_SB_SEI_MASTERING_DISPLAY_COLOR_VOLUME; type < VPI_SB_TYPE_MAX; ++type) {
      if (vpifrm_get_sidedata_count_by_type(src, static_cast<VpiSidedataType>(type)) > 0) {
        needs_hdr = true;
        ++count;
        break;
      }
    }
  } else {
    for (uint32_t type = VPI_SB_ML_ROI; type < VPI_SB_TYPE_MAX; ++type) {
      count += vpifrm_get_sidedata_count_by_type(src, static_cast<VpiSidedataType>(type));
    }
  }

  XmaFrameSideData** sidedata_array = new XmaFrameSideData*[count + 1];
  if (!sidedata_array) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopySideData() Out of memory!");
  }
  memset(sidedata_array, 0, sizeof(XmaFrameSideData*) * (count + 1));
  m_frame->side_data = sidedata_array;

  uint32_t       index = 0;
  IntSideDataPtr sd;
  // only add XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS if something is changing
  if (has_dyn_enc_params) {
    if (api_version < XMA_API_VERSION_1_1_2) {
      sd = InternalSideData::AllocSideData(m_sf, m_log, XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS, XMA_HOST_BUFFER_TYPE, sizeof(XmaDynamicEncParams));
      if (!sd) {
        THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopySideData() Out of memory!");
      }
      memcpy(sd->GetHostPtr(), &src->dynamic_params, sizeof(XmaDynamicEncParams));
      sd->GetPrivateInfo()->IncFramesHolding();
      sidedata_array[index++] = sd->GetSideData();
    } else {
      sd = InternalSideData::AllocSideData(m_sf, m_log, XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS, XMA_HOST_BUFFER_TYPE, sizeof(XmaDynamicEncParams_v2));
      if (!sd) {
        THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopySideData() Out of memory!");
      }
      memcpy(sd->GetHostPtr(), &src->dynamic_params, sizeof(XmaDynamicEncParams_v2));
      sd->GetPrivateInfo()->IncFramesHolding();
      sidedata_array[index++] = sd->GetSideData();
    }
  }

  if (api_version < XMA_API_VERSION_1_1) {
    if (vpifrm_get_sidedata_count_by_type(src, VPI_SB_VUI_PARAMS) > 0) {
      VpiSidedataHandle vpi_sidedata = src->vpi_sidedata[VPI_SB_VUI_PARAMS][0];
      vpi_sidedata_ref(vpi_sidedata);
      side_data_buffer_t buffer = side_data_buffer_t(vpi_sidedata, [](void* buf) { vpi_sidedata_unref(buf); });
      sd                        = InternalSideData::AllocSideData(m_sf, m_log, buffer);
      if (!sd) {
        THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopySideData() Out of memory!");
      }
      sd->GetPrivateInfo()->IncFramesHolding();
      sidedata_array[index++] = sd->GetSideData();
    }

    if (needs_hdr) {
      sd = InternalSideData::AllocSideData(m_sf, m_log, XMA_FRAME_SIDE_DATA_HDR10_PARAMS, XMA_HOST_BUFFER_TYPE, sizeof(XmaHdrParams));
      if (!sd) {
        THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopySideData() Out of memory!");
      }
      XmaHdrParams* hdr_params = reinterpret_cast<XmaHdrParams*>(sd->GetHostPtr());

      if (src->vpi_sidedata[VPI_SB_SEI_MASTERING_DISPLAY_COLOR_VOLUME][0]) {
        VpiSidedataHandle vpi_sidedata = src->vpi_sidedata[VPI_SB_SEI_MASTERING_DISPLAY_COLOR_VOLUME][0];
        void*             src_data     = vpi_get_sidedata_vaddr(vpi_sidedata);
        memcpy(&hdr_params->mdcv, src_data, vpi_get_sidedata_size(vpi_sidedata));
      }

      if (src->vpi_sidedata[VPI_SB_SEI_CONTENT_LIGHT_LEVEL][0]) {
        VpiSidedataHandle vpi_sidedata = src->vpi_sidedata[VPI_SB_SEI_CONTENT_LIGHT_LEVEL][0];
        void*             src_data     = vpi_get_sidedata_vaddr(vpi_sidedata);
        memcpy(&hdr_params->cll, src_data, vpi_get_sidedata_size(vpi_sidedata));
      }

      if (src->vpi_sidedata[VPI_SB_SEI_ATC][0]) {
        VpiSidedataHandle vpi_sidedata = src->vpi_sidedata[VPI_SB_SEI_ATC][0];
        void*             src_data     = vpi_get_sidedata_vaddr(vpi_sidedata);
        memcpy(&hdr_params->atc, src_data, vpi_get_sidedata_size(vpi_sidedata));
      }

      if (src->vpi_sidedata[VPI_SB_SEI_RAW_T35_DATA][0]) {
        hdr_params->t35 = new XmaT35Params;
        memset(hdr_params->t35, 0, sizeof(XmaT35Params));
        for (uint32_t i = 0; (i < VPI_MAX_SB_NUM) && (src->vpi_sidedata[VPI_SB_SEI_RAW_T35_DATA][i]); ++i) {
          VpiSidedataHandle vpi_sidedata                                       = src->vpi_sidedata[VPI_SB_SEI_RAW_T35_DATA][i];
          VpiRawT35Data*    raw_t35                                            = reinterpret_cast<VpiRawT35Data*>(vpi_get_sidedata_vaddr(vpi_sidedata));
          int32_t           size                                               = raw_t35->payload_byte_length;
          hdr_params->t35->payload_type[hdr_params->t35->payload_count]        = VPE_SEI_USER_DATA_REGISTERED_ITU_T_T35;
          hdr_params->t35->payload_byte_length[hdr_params->t35->payload_count] = size;
          hdr_params->t35->payload_byte_data[hdr_params->t35->payload_count]   = new uint8_t[size];
          memcpy(hdr_params->t35->payload_byte_data[hdr_params->t35->payload_count], raw_t35->payload_byte_data, size);
          if (size >= 4) {
            if ((raw_t35->payload_byte_data[0] == 0x00) && (raw_t35->payload_byte_data[1] == 0x3B) && (raw_t35->payload_byte_data[2] == 0x00) &&
                (raw_t35->payload_byte_data[3] == 0x00)) {
              vpi_sidedata                        = src->vpi_sidedata[VPI_SB_SEI_DOLBY_VISION][0];
              XmaSMPTE_ST2094_10_v2* dolby_vision = reinterpret_cast<XmaSMPTE_ST2094_10_v2*>(vpi_get_sidedata_vaddr(vpi_sidedata));
              memcpy(&hdr_params->t35->meta[hdr_params->t35->payload_count].st2094_10, &dolby_vision->st2094_10, vpi_get_sidedata_size(vpi_sidedata));
              hdr_params->t35->country_code[hdr_params->t35->payload_count]          = dolby_vision->country_code;
              hdr_params->t35->extended_country_code[hdr_params->t35->payload_count] = dolby_vision->extended_country_code;
            } else if ((raw_t35->payload_byte_data[0] == 0x00) && (raw_t35->payload_byte_data[1] == 0x3C) && (raw_t35->payload_byte_data[2] == 0x00) &&
                (raw_t35->payload_byte_data[3] == 0x01)) {
              vpi_sidedata                      = src->vpi_sidedata[VPI_SB_SEI_HDR10_PLUS][0];
              XmaSMPTE_ST2094_40_v2* hdr10_plus = reinterpret_cast<XmaSMPTE_ST2094_40_v2*>(vpi_get_sidedata_vaddr(vpi_sidedata));
              memcpy(&hdr_params->t35->meta[hdr_params->t35->payload_count].st2094_40, &hdr10_plus->st2094_40, vpi_get_sidedata_size(vpi_sidedata));
              hdr_params->t35->country_code[hdr_params->t35->payload_count]          = hdr10_plus->country_code;
              hdr_params->t35->extended_country_code[hdr_params->t35->payload_count] = hdr10_plus->extended_country_code;
            }
          }
          ++hdr_params->t35->payload_count;
        }
      }
      sd->GetPrivateInfo()->IncFramesHolding();
      sidedata_array[index++] = sd->GetSideData();
    }
  } else {
    // add all other side data
    for (uint32_t type = VPI_SB_ML_ROI; type < VPI_SB_TYPE_MAX; ++type) {
      for (uint32_t i = 0; (i < VPI_MAX_SB_NUM) && (src->vpi_sidedata[type][i]); ++i) {
        VpiSidedataHandle vpi_sidedata = src->vpi_sidedata[type][i];
        vpi_sidedata_ref(vpi_sidedata);
        side_data_buffer_t buffer = side_data_buffer_t(vpi_sidedata, [](void* buf) { vpi_sidedata_unref(buf); });
        sd                        = InternalSideData::AllocSideData(m_sf, m_log, buffer);
        if (!sd) {
          THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopySideData() Out of memory!");
        }
        sd->GetPrivateInfo()->IncFramesHolding();
        sidedata_array[index++] = sd->GetSideData();
      }
    }
  }
}

///////////////////////////////////////////////////////////////////////////////
VpiBufferPool* GetVpiPool(LogPtr log, FramePtr frame) {
  if (!frame) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "GetVpiPool() Invalid parameter!");
  }

  FramePropertiesPtr props = frame->GetProps();

  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "GetVpiPool(format=%s, sw_format=%s, size=%dx%d)", InternalFrameProperties::GetPixFmtName(props->GetFormat()),
      InternalFrameProperties::GetPixFmtName(props->GetSwFormat()), props->GetWidth(), props->GetHeight());
  VpiBufferPool* pool      = nullptr;
  IntFramePtr    int_frame = std::dynamic_pointer_cast<InternalFrame>(frame);
  if (int_frame) {
    XmaFrame* xma_frame = int_frame->GetFrame();
    if (xma_frame) {
      LOCK_XMA_FRAME(log, "GetVpiPool", XMA_FRAME, xma_frame);
      PrivateFrameInfo* frame_info = static_cast<PrivateFrameInfo*>(xma_frame->private_do_not_touch);
      if (frame_info && frame_info->IsValid()) {
        pool = frame_info->GetBufferPool();
      }
    }
  } else {
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "GetVpiPool() frame not valid!");
  }
  if (!pool) {
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "GetVpiPool() Unable to find buffer pool!");
  }
  return pool;
}

///////////////////////////////////////////////////////////////////////////////
int32_t AddFrameToVpiFrm(IntSessionFactoryPtr sf, LogPtr log, FramePtr frame, VpiFrm* pic) {
  if (!frame) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "AddFrameToVpiFrm() Invalid parameter!");
  }

  IntFramePtr int_frame = std::dynamic_pointer_cast<InternalFrame>(frame);
  if (!int_frame) {
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "AddFrameToVpiFrm() frame not valid!");
  }

  LOCK_XMA_FRAME(log, "AddFrameToVpiFrm", XMA_FRAME, int_frame->m_frame);

  if (!pic) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "AddFrameToVpiFrm() Invalid parameter!");
  }

  FramePropertiesPtr props = frame->GetProps();

  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "AddFrameToVpiFrm(format=%s, sw_format=%s, size=%dx%d)", InternalFrameProperties::GetPixFmtName(props->GetFormat()),
      InternalFrameProperties::GetPixFmtName(props->GetSwFormat()), props->GetWidth(), props->GetHeight());

  if (((props->GetFormat() == XMA_VPE_FMT_TYPE) && (frame->GetBufferType() == XMA_HOST_BUFFER_TYPE)) ||
      ((props->GetFormat() != XMA_VPE_FMT_TYPE) && (frame->GetBufferType() != XMA_HOST_BUFFER_TYPE)) || (frame->GetBufferType() == NO_BUFFER)) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "AddFrameToVpiFrm() Inconsistent buffer types!");
  }

  int32_t planes = props->GetNumPlanes();
  int32_t plane;

  int32_t ch;
  if (!pic->valid) {
    pic->valid     = 1;
    pic->width     = props->GetWidth();
    pic->height    = props->GetHeight();
    pic->poc       = -1;
    pic->format    = static_cast<VpiFmt>(props->GetFormat());
    pic->key_frame = frame->IsIdr();
    pic->pts       = frame->GetPts();
    pic->dts       = -1;
    ch             = 0;
  } else {
    for (ch = 0; (ch < VPE_CHANNEL_NUM); ++ch) {
      if (!pic->ch_info[ch].enabled) {
        break;
      } else if ((props->GetWidth() == pic->ch_info[ch].width) && (props->GetHeight() == pic->ch_info[ch].height)) {
        THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME,
            "AddFrameToVpiFrm() Unable to add frame to pic, another frame with the same dimensions(%dx%d) has already been added!", props->GetWidth(),
            props->GetHeight());
      }
    }

    if (ch == VPE_CHANNEL_NUM) {
      THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "AddFrameToVpiFrm() No available free channels!");
    }
  }

  pic->ch_info[ch].enabled = 1;
  pic->ch_info[ch].width   = props->GetWidth();
  pic->ch_info[ch].height  = props->GetHeight();
  pic->ch_info[ch].walign  = XMA_FRAME_W_ALIGN;
  pic->ch_info[ch].halign  = XMA_FRAME_H_ALIGN;
  pic->ch_info[ch].flag    = props->GetFlags() | VPI_FRAME_NEW_FLAG;
  for (plane = 0; plane < planes; ++plane) {
    pic->ch_info[ch].stride_size[plane] = props->GetPlaneStride(plane);
  }

  pic->ch_info[ch].format    = static_cast<VpiFmt>(props->GetFormat());
  pic->ch_info[ch].sw_format = static_cast<VpiFmt>(props->GetSwFormat());

  switch (frame->GetBufferType()) {
  case XMA_HOST_BUFFER_TYPE:
  case XMA_DEVICE_ONLY_BUFFER_TYPE: {
  case XMA_DEVICE_BUFFER_TYPE:
    if (ch == 0) {
      buffer_t vpi_buf = sf->GetBuffer(0);
      vpi_buffer_set_data(vpi_buf.get(), pic);
      vpi_buffer_set_free_cb(vpi_buf.get(), xma_pic_consume);
      vpi_buffer_set_opaque(vpi_buf.get(), nullptr);
      pic->vpi_picinfo_buf = vpi_buf.release();
    }

    PrivateVpiFrmInfo* info = static_cast<PrivateVpiFrmInfo*>(pic->opaque);
    if (info && info->IsValid()) {
      info->AddFrame(frame, pic, ch);
    } else {
      info = new PrivateVpiFrmInfo(sf, log);
      info->AddFrame(frame, pic, ch);
      VpiBuffer* vpi_buf = static_cast<VpiBuffer*>(pic->vpi_picinfo_buf);
      info->SetFreeBuf(vpi_buf->freebuf);
      vpi_buf->freebuf = xma_free_reference_info;
      info->SetPool(static_cast<VpiBufferPool*>(vpi_buf->pool));
      vpi_buf->pool = GetVpiPool(log, frame);
      pic->opaque   = info;
    }
    vpi_frame_ref(pic);
  } break;
  case NO_BUFFER:
    break;
  default:
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "AddFrameToVpiFrm() Unknown buffer type!");
  }

  AddSideDataToVpiFrm(sf, log, frame, pic);

  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "AddFrameToVpiFrm() - Returning");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t CopyFrame(LogPtr log, XmaFrame* dst, FramePtr src) {
  if (!dst || (dst->data[0].buffer_type != NO_BUFFER) || !src) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "CopyFrame() Invalid parameter!");
  }

  IntFramePtr src_frame = std::dynamic_pointer_cast<InternalFrame>(src);
  if (!src_frame) {
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "CopyFrame() source not valid!");
  }

  LOCK_XMA_FRAME(log, "CopyFrame", XMA_FRAME, src_frame->m_frame);

  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "CopyFrame(format=%s, sw_format=%s, size=%dx%d)", InternalFrameProperties::GetPixFmtName(src->GetProps()->GetFormat()),
      InternalFrameProperties::GetPixFmtName(src->GetProps()->GetSwFormat()), src->GetProps()->GetWidth(), src->GetProps()->GetHeight());

  PrivateFrameInfo* dst_info = static_cast<PrivateFrameInfo*>(dst->private_do_not_touch);
  if ((!dst_info) || (!dst_info->IsValid())) {
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "CopyFrame() Invalid private info!");
  }

  memcpy(dst, src_frame->m_frame.get(), sizeof(XmaFrame));

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(src_frame->m_sf->m_mutex);
  if (src_frame->m_sf->m_frames.remove(src_frame->m_frame.get()) != 1) {
    log->LogMsg(XMA_ERROR_LOG, XMA_FRAME, "XMA Memory check Error! Frame %p removed twice?", src_frame->m_frame.get());
  }
  lock.unlock();
#endif

  IntFramePtr dst_frame = dst_info->GetSelf();
  src_frame->m_frame    = std::move(dst_frame->m_frame);
  delete dst_info;

  PrivateFrameInfo* src_info = static_cast<PrivateFrameInfo*>(src_frame->m_frame->private_do_not_touch);
  if ((!src_info) || (!src_info->IsValid())) {
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "CopyFrame() Invalid private info!");
  }
  src_info->SetSelf(src_frame);

  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "CopyFrame() - Returning");
  return XMA_SUCCESS;
}

}; // Namespace xma
