// Copyright (C) 2023, AMD Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include "SDBufferPool.h"
#include "InternalSideData.h"
#include "vpi_api.h"
#include "vpi_sidedata.h"

namespace xma {

#define XMA_SD_BUFFER_POOL "SDBufferPool"
#define NEXT_MULTIPLE(value, n) (((value) + (n) -1) & ~((n) -1))

//////////////////////////////////////////////////////////////////////////////
SDBufferPool::SDBufferPool(VpiMemHandle mem, LogPtr log, enum XmaFrameSideDataType type, uint32_t num, uint32_t size, uint32_t flag) : m_type(type), m_log(log) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SD_BUFFER_POOL, "SDBufferPool::SDBufferPool(num = %u, size = %u)", num, size);
  VpiBufferPool* new_pool = nullptr;
  VpiRet         ret      = vpi_buffer_pool_create(mem, num, NEXT_MULTIPLE(size, 4096), flag, false, 0, &new_pool);
  if (ret) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SD_BUFFER_POOL, "SDBufferPool::SDBufferPool() Failed to create side data buffer pool: %s!", vpi_error_str(ret));
  }
  m_pool = pool_t(new_pool, [](VpiBufferPool* pool) { vpi_buffer_pool_unref(pool); });
  if (!m_pool) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SD_BUFFER_POOL, "SDBufferPool::SDBufferPool() Out of memory!");
  }

  InitData();

  m_pool->opaque = this;

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SD_BUFFER_POOL, "SDBufferPool::SDBufferPool() - returning");
}

///////////////////////////////////////////////////////////////////////////////
SDBufferPool::~SDBufferPool() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SD_BUFFER_POOL, "SDBufferPool::~SDBufferPool()");

  m_pool->opaque = nullptr;
}

///////////////////////////////////////////////////////////////////////////////
int32_t SDBufferPool::SetLog(LogPtr log) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SD_BUFFER_POOL, "SDBufferPool::SetLog()");
  lock_t guard(m_mutex);
  m_log = log;
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SD_BUFFER_POOL, "SDBufferPool::SetLog() - Returning");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
side_data_buffer_t SDBufferPool::GetBuffer(uint32_t size) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SD_BUFFER_POOL, "SDBufferPool::GetBuffer()");
  lock_t guard(m_mutex);

  InitData();

  VpiBuffer* buf = nullptr;
  while (!buf) {
    buf = vpi_buffer_pool_alloc_item(m_pool.get());
    if (!buf || !buf->m_handle) {
      if (m_pool->num == MAX_POOL_ITEM_NUM) {
        THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SD_BUFFER_POOL, "SDBufferPool::GetBuffer() Unable to auto-extend buffer pool!");
      }
      guard.unlock();
      Extend(SD_BUFFER_POOL_AUTO_EXTEND_SIZE);
      guard.lock();
    }
  }

  uint32_t bufsize = vpi_buffer_get_size(buf);
  if (bufsize < size) {
    uint32_t newsize = NEXT_MULTIPLE(size, 4096);
    if (vpi_buffer_pool_extend_mem_size(m_pool.get(), buf, newsize, 0)) {
      THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SD_BUFFER_POOL, "SDBufferPool::GetBuffer() Out of memory!");
    }
  }

  if (!buf->data) {
    buf->data = vpi_memory_malloc(sizeof(VpiSidedata));
    if (!buf->data) {
      THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SD_BUFFER_POOL, "SDBufferPool::GetBuffer() Out of memory!");
    }
    memset(buf->data, 0, sizeof(VpiSidedata));
  }

  VpiSidedata* sd = reinterpret_cast<VpiSidedata*>(buf->data);
  sd->data        = buf;
  sd->type        = InternalSideData::GetVpiTypeFromXmaType(m_type);
  sd->size        = size;
  sd->pool        = m_pool.get();
  sd->extension   = nullptr;
  vpi_atomic_init(&buf->refcnt, 1);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SD_BUFFER_POOL, "SDBufferPool::GetBuffer() - returning");
  return side_data_buffer_t(sd, [&](void* sd_buf) {
    VpiSidedata* vpi_sd = reinterpret_cast<VpiSidedata*>(sd_buf);
    vpi_buffer_unref(vpi_sd->data);
  });
}

///////////////////////////////////////////////////////////////////////////////
int32_t SDBufferPool::Extend(uint32_t num) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SD_BUFFER_POOL, "SDBufferPool::Extend(num = %u)", num);
  lock_t guard(m_mutex);
  if (vpi_buffer_pool_extend(m_pool.get(), num)) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SD_BUFFER_POOL, "SDBufferPool::Extend() Unable to extend buffer pool!");
  }

  InitData();

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SD_BUFFER_POOL, "SDBufferPool::Extend() - returning");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
void SDBufferPool::InitData() {
  for (uint32_t i = m_size; i < m_pool->num; ++i) {
    VpiBuffer* buf = vpi_buffer_pool_get_list_item(m_pool.get(), i);
    vpi_buffer_set_free_cb(buf, [](void* opaque, void* data, void* /*buff*/) {
      if (!data) {
        return;
      }
      VpiSidedata*   vpi_sd   = static_cast<VpiSidedata*>(data);
      VpiBuffer*     vpi_buf  = vpi_sd->data;
      VpiBufferPool* vpi_pool = static_cast<VpiBufferPool*>(opaque);
      SDBufferPool*  xma_pool = static_cast<SDBufferPool*>(vpi_pool->opaque);

      switch (xma_pool->m_type) {
      case XMA_FRAME_SIDE_DATA_HDR10_PARAMS: {
        XmaHdrParams* hdr_params = reinterpret_cast<XmaHdrParams*>(vpi_buffer_get_vaddr(vpi_buf, 0, 0));
        if (hdr_params && hdr_params->t35) {
          for (uint8_t index = 0; index < hdr_params->t35->payload_count; ++index) {
            if (hdr_params->t35->payload_byte_data[index]) {
              delete[] hdr_params->t35->payload_byte_data[index];
            }
          }
          delete hdr_params->t35;
          hdr_params->t35 = nullptr;
        }
        break;
      }
      case XMA_FRAME_SIDE_DATA_SEI_RAW_T35_DATA: {
        XmaRawT35Data* raw_t35 = reinterpret_cast<XmaRawT35Data*>(vpi_buffer_get_vaddr(vpi_buf, 0, 0));
        if (raw_t35 && raw_t35->payload_byte_data) {
          delete[] raw_t35->payload_byte_data;
          raw_t35->payload_byte_data = nullptr;
        }
        break;
      }
      default:
        break;
      }

      vpi_sd->type      = VPI_SB_NONE;
      vpi_sd->size      = 0;
      vpi_sd->data      = nullptr;
      vpi_sd->pool      = nullptr;
      vpi_sd->extension = nullptr;

      vpi_buffer_pool_free_item(vpi_pool, vpi_buf);
    });
  }
  m_size = m_pool->num;
}

} // namespace xma
